from __future__ import annotations
from abc import abstractmethod
from typing import TYPE_CHECKING, Any, Union
import numpy as np

if TYPE_CHECKING:
    from ._base import XTensor
    from .typing import HasDimName



class MetaTensorSlice:
    def __init__(self, dimname: str):
        self._dimname = dimname

    def __getitem__(self, slc: Union[int, slice]) -> TensorIndexer:
        if isinstance(slc, slice):
            return TensorSlice(self.dimname, slc.start, slc.stop, slc.step)
        else:
            return SingleIndex(self.dimname, slc)
        
    @property
    def dimname(self) -> str: return self._dimname

    def __get_dimname__(self) -> str:
        return self.dimname



class TensorIndexer:
    @abstractmethod
    def index(self, X: XTensor) -> XTensor: ...


class TensorSlice(TensorIndexer):
    def __init__(self, dimname: str, start, stop, step):
        self._dimname = dimname
        self._slice = slice(start, stop, step)

    def index(self, X: XTensor) -> XTensor:
        Y = X.viewcopy()
        axis = Y.get_axis(self._dimname)

        coord = X.get_coord(axis)
        coord = None if coord is None else coord[self._slice]

        slices = [slice(None, None, None) for _ in range(X.rank)]
        slices[axis] = self._slice

        Y.data = X.data[tuple(slices)]

        Y.set_coord(axis, coord=coord)
        return Y


class SingleIndex(TensorIndexer):
    def __init__(self, dimname: str, ind: int):
        self._dimname = dimname
        self.ind = ind

    def index(self, X: XTensor) -> XTensor:
        return X.get(self._dimname, self.ind)






from typing import Literal, Optional, Sequence, Tuple, TypeVar, cast

from xarray import DataArray
from ..typing import NDArray, DimLike, DimsLike, AxisDimPair
import numpy as np

from ._base import is_named

from ._conv import to_dataarray, to_ndarray



def get_dims(x: NDArray) -> Tuple[str|None,...]:
    '''
    Return a list of dimnesion names for the given tensor
    Unnamed dimensions yield None
    '''

    if isinstance(x, DataArray):
        _dims = []
        for dim in x.dims:
            dim = cast(str, dim)

            if not is_named(dim):
                _dims.append(None)
            else:
                _dims.append(dim)

        return tuple(_dims)

    return tuple(None for _ in x.shape)


T = TypeVar('T')
def find_index(objs: Sequence[T], target: T) -> int:
    '''
    Given a tuple of objects, find the index where target occurs. If there
    is not exactly one occurrence, an IndexError is raised.
    '''
    _dims = np.array(objs)
    
    index = np.where(_dims == target)[0]

    if len(index) == 0:
        raise IndexError(f'{target} is not found')

    if len(index) > 1:
        raise IndexError(f'{target} is ambiguous')

    index = index.item()
    return index


def get_axis_from_dim(x: NDArray, dim: str, fallback_axis: Optional[int]=None) -> int:
    '''
    Given a single dimension (string) and an array x, try locating the
    corresponding axis (int). If not found, the fallback is returned if
    provided, else an IndexError is raised.
    '''


    try:
        if not is_named(dim): raise IndexError(f'{dim} is unnamed')

        if isinstance(x, DataArray):
            return find_index(cast(Tuple[str,...], x.dims), dim)
        raise IndexError(f'trying to locate dimension {dim} but array/tensor is not named')

    except IndexError as e:
        if fallback_axis is not None:
            return fallback_axis
        
        raise IndexError(*e.args)


def get_axes(x: NDArray, dim: DimLike|DimsLike|None) -> Tuple[int,...]:
    '''
    Given dimensions (in strings) or axes (in integers), return a list of
    integers corresponding to the dimensions/axes

    Parameters:
        x: Input array
        dim: dimensions specifications, e.g.:
            2
            'N'
            (2, 'N') - 'N', or 2 if 'N' is not found
            (2, 3, 4)
            ('N', 'C', 'H')
            ((2, 'N'), (3, 'C'), (4, 'H'))

    '''

    if dim is None:
        return tuple([i for i in range(len(x.shape))])

    if isinstance(dim, int):
        return (dim,)

    if isinstance(dim, str):
        return (get_axis_from_dim(x, dim),)
    
    if isinstance(dim, tuple):
        if isinstance(dim[0], int):
            if isinstance(dim[1], str):
                return (get_axis_from_dim(x, dim[1], fallback_axis=dim[0]),)
            
            return cast(Tuple[int,...], dim)

        if isinstance(dim[0], str):
            dim = cast(Tuple[str,...], dim)
            return tuple(get_axis_from_dim(x, _dim) for _dim in dim)

        if isinstance(dim[0], tuple):
            dim = cast(Tuple[AxisDimPair,...], dim)
            return tuple(get_axis_from_dim(x, _dim[1], fallback_axis=_dim[0]) for _dim in dim)

    raise ValueError('Invalid argument combination')


def strip_dims(olddims: Tuple[str, ...], axis: Tuple[int, ...]) -> Tuple[str,...]:
    '''
    (For named dimensions) Remove dimensions corresponding to [axis] from named
    dimensions.
    '''
    newdims = list(olddims)
    stripped_dims = [olddims[i] for i in axis] 

    for sd in stripped_dims: newdims.remove(sd)

    return tuple(newdims)


def rename(x: NDArray, dim: DimLike, newname: str) -> DataArray:
    '''
    Rename a dimension. If the original array is unnamed, a named array will be
    returned.
    '''

    if not is_named(newname): raise ValueError(f'Invalid dimension name: {newname}')

    x_ = to_dataarray(x)
    _axis = get_axes(x_, dim)[0]
    _dim = get_dims(x)[_axis]
    return x_.rename({_dim: newname})


def new_axes(x: NDArray, *dims: str, position: Literal['before','after']='after') -> DataArray:
    _x = to_dataarray(x)
    if position == 'after':
        for dim in dims:
            _x = new_axis(_x, dim, position='after')
        return _x

    if position == 'before':
        for dim in dims[::-1]:
            _x = new_axis(_x, dim, position='before')
        return _x

    raise ValueError(f'Invalid position: {position}')


def new_axis(x: NDArray, dim: str, position: Literal['before','after']='after') -> DataArray:
    _x = to_dataarray(x)
    dims = list(_x.dims)
    _x = to_ndarray(_x)

    if position == 'before':
        _x = _x[None,...]
        _x = DataArray(_x, dims=['__dims_appended_tmp'] + dims)
        _x = rename(_x, 0, dim)
        return _x

    if position == 'after':
        _x = _x[...,None]
        _x = DataArray(_x, dims=dims + ['__dims_appended_tmp'])
        _x = rename(_x, -1, dim)
        return _x

    raise ValueError(f'Invalid position: {position}')






from ._misc import where, softmax, get_rank

from ._binary_op import (
    _add as add,
    _divide as divide,
    _multiply as multiply,

    _greater as greater,
    _greater_equal as greater_equal,
    _less as less,
    _less_equal as less_equal,
    _equal as equal,
)

from ._reduc import (
    _all as all,
    _any as any,
    _max as max,
    _min as min,
    _nanmax as nanmax,
    _nanmin as nanmin,

    _mean as mean,
    _nanmean as nanmean,
    _nanstd as nanstd,
    _nansum as nansum,
    _std as std,
    _sum as sum,
)


from ._arg import (
    _argmax as argmax,
    _argmin as argmin,
    _nanargmax as nanargmax,
    _nanargmin as nanargmin,
    _coordmax as coordmax,
    _coordmin as coordmin,
    _nancoordmax as nancoordmax,
    _nancoordmin as nancoordmin
)

from ._args import (argsmin, argsmax, nanargsmax, nanargsmin, coordsmin, coordsmax)

from ._reduc_2to1 import diagonal

from ._ufuncs import cos, cosh, exp, log, log2, log10, sigmoid, sin, sinh, tan, tanh 







from __future__ import annotations
'''
Generalization:
    These decorators take XTensor-only functions and make them compatible with
    any object that supports the __array__() protocol, which is equivalent to
    an XTensor with no named dimensions.

'''
from functools import wraps
from typing import TYPE_CHECKING, Any, Callable, TypeVar
from typing_extensions import ParamSpec, Concatenate
import typing

from ._base import to_xtensor


if TYPE_CHECKING:
    from .._base import XTensor
    from ..typing import TensorLike

O = ParamSpec('O')

T = TypeVar('T')
U = TypeVar('U')
V = TypeVar('V')
R = TypeVar('R')


def generalize_at_0(
        func: Callable[Concatenate[XTensor, O], R]
    ) -> Callable[Concatenate[TensorLike, O], R]:
    '''
    Generalize the first argument of a function from accepting XTensor to
    TensorLike objects
    '''
    @wraps(func)
    def wrapped(X: TensorLike, *args: O.args, **kwargs: O.kwargs) -> R:
        return func(to_xtensor(X), *args, **kwargs)
    return wrapped

def generalize_at_1(
        func: Callable[Concatenate[T, XTensor, O], R]
    ) -> Callable[Concatenate[T, TensorLike, O], R]:
    # Function_2Args[T, TensorLike, O, R]:
    '''Generalize the second argument of a function'''
    @wraps(func)
    def wrapped(arg0: T, X: TensorLike, *args: O.args, **kwargs: O.kwargs) -> R:
        return func(arg0, to_xtensor(X), *args, **kwargs)
    return wrapped


def generalize_at_2(
        func: Callable[Concatenate[T, U, XTensor, O], R]
    ) -> Callable[Concatenate[T, U, TensorLike, O], R]:
    '''Generalize the second argument of a function'''
    @wraps(func)
    def wrapped(arg0: T, arg1: U, X: TensorLike, *args: O.args, **kwargs: O.kwargs) -> R:
        return func(arg0, arg1, to_xtensor(X), *args, **kwargs)
    return wrapped


def generalize_at_3(
        func: Callable[Concatenate[T, U, V, XTensor, O], R]
    ) -> Callable[Concatenate[T, U, V, TensorLike, O], R]:
    '''Generalize the second argument of a function'''
    @wraps(func)
    def wrapped(arg0: T, arg1: U, arg2: V, X: TensorLike, *args: O.args, **kwargs: O.kwargs) -> R:
        return func(arg0, arg1, arg2, to_xtensor(X), *args, **kwargs)
    return wrapped




from ._base import to_xtensor

from ._axes import (
        permute, newdims, align, shapes_broadcastable,
        permutation_well_defined, shape, stack, rank, index)

from ._dims import (
        mergedims, flatten, dimsfirst, dimslast,
        name_dim_if_absent, dims)

from ._coords import mergecoords, coords_same

from ._generalize import generalize_at_0, generalize_at_1, generalize_at_2, generalize_at_3


from ._misc import strip 





from typing import Protocol


import numpy as np
import numpy.typing as npt

from .. import tensor as xtt

'''
Arg functions:
    For single-dimension reduction functions whose return values are array
    **indices** or **coordinates**
    E.g. argmax
'''

class _np_arg_func(Protocol):
    def __call__(self, a: np.ndarray, axis: int) -> npt.NDArray[np.int_]: ...

class ArgFunction(Protocol):
    def __call__(self, x: xtt.TensorLike, /, dim: xtt.DimLike) -> xtt.XTensor: ...

class CoordFunction(Protocol):
    def __call__(self, x: xtt.TensorLike, /, dim: xtt.DimLike, *, 
            use_index_if_no_coord: bool=False) -> xtt.XTensor: ...


def _reduction_factory(_np_func: _np_arg_func) -> ArgFunction:
    @xtt.generalize_at_0
    def _reduce(X: xtt.XTensor, /, dim: xtt.DimLike) -> xtt.XTensor:
        axis = X.get_axis(dim)
        return xtt.XTensor(
                _np_func(X.data, axis=axis),
                dims=xtt.strip(X.dims, [axis]),
                coords=xtt.strip(X.coords, [axis]))
    return _reduce

_argmax = _reduction_factory(np.argmax)
_argmin = _reduction_factory(np.argmin)

_nanargmax = _reduction_factory(np.nanargmax)
_nanargmin = _reduction_factory(np.nanargmin)


def _coord_reduc_factory(_func: ArgFunction) -> CoordFunction:
    @xtt.generalize_at_0
    def _reduce(X: xtt.XTensor, /, dim: xtt.DimLike, *, use_index_if_no_coord: bool=False) -> xtt.XTensor:
        '''
            Return the coordinate on [dim] that maximizes/minimizes x If dimension [dim] does
            not have coordinates, this is equivalent to (nan)argmax/argmin
        '''
        args = _func(X, dim)
        axis = X.get_axis(dim)
        coord = X.coords[axis]

        if coord is None: 
            if use_index_if_no_coord:
                coord_r = np.arange(X.shape[axis])[args]
            else:
                raise ValueError(f'Tensor has no coordinates on axis {axis}')
        else:
            coord_r = coord[args]

        return xtt.XTensor(coord_r, dims=xtt.strip(X.dims, [axis]), coords=xtt.strip(X.coords, [axis]))
    return _reduce


_coordmax = _coord_reduc_factory(_argmax)
_coordmin = _coord_reduc_factory(_argmin)

_nancoordmax = _coord_reduc_factory(_nanargmax)
_nancoordmin = _coord_reduc_factory(_nanargmin)





'''
# 1. Broadcasting

This module tries to address the problem of broadcasting two tensors.

Traditionally, two unnamed tensors are broadcast by aligning them from the right, e.g.

    X (64, 32, 8, 5)

    Y     (32, 8, 1)

are broadcastable, while

    X (64, 32, 8, 5)

    Y     (64, 32, 8)

are not.


The problem can be reduced to finding a consistent policy to match dimensions
of two given tensors. They are then broadcastable if and only if the matched
dimensions satisfy one of the following conditions:

    (a) they have the same length
    (b) one of them is singleton (length 1)


In general, two tensors are broadcastable if and only if:
    
    **there exists a matching scheme such that one of the tensors has its
    axes completely matched to those of the other one**


Formally, let X and Y be two tensors of shapes (x1, ..., xM) and (y1, ..., yN),
WLOG assume M > N, then they are broadcastable if and only if there exists a
matching

    (i1, j1), (i2, j2), ..., (iN, jN)

    such that (xip = yip) or (xip = 1) or (yip = 1)

    for p = 1, ..., N


# 2. Broadcasters

A *broadcaster* should:

    1. Permute the axes of X and Y, adding singleton dimensions if necessary,
       so that they are broadcastable by the default reverse-index policy

    2. 



# 3. Named Tensors

This is further complicated when named dimensions are introduced. In general,
one may want to use named tensors without having to name all the dimensions,
consider for example two partially named tensors

    img: 
        shape  (10, 3,       256, 384)
        axis    0   1        2    3
        name    ?   CHANNEL  H    W


    lbl:

        shape  (10, 256, 384)
        axis    0   1    2
        name    ?   H    W

The desired broadcasting here should be 
    
    (img axis, lbl axis): (0, 0) (2, 1) (3, 2)

    output: (10, 3, 256, 384)



Another example:

    lbl_truth:
        shape   (20, 512, 512)
        axis     0   1    2
        name     ?   H    W

    lbl_prob:
        shape   (20, 3,     512, 512)
        axis     0   1      2    3
        name     ?   CLASS  H    W

    lbl_pred:
        shape   (20, 1,     17,    15,    512,  512)
        axis     ?   SCALE1 SCALE2 SCALE3 H     W


# 4. Principles
    
    (a) The result of broadcasting two tensors X and Y should have the same
        number of dimensions as the tensor with more dimensions

    (b) If X and Y has dimensions with the same name, they should be broadcast
        together



'''

from ._broadcast import broadcast, vanilla_broadcaster, template_broadcaster, unilateral_broadcaster, cast

from ._dimcast import castdim, unilateral_dimcast

from ._template import Template, AxisSelector, IndexSelector, DimNameSelector
from ._types import Broadcaster, Dimcaster




from ._base import XTensor


from ._decors import promote_binary_operator, promote_ternary_operator

from ._slice import TensorSlice, MetaTensorSlice


from .broadcast import *

from .basic_utils import *

from .typing import *



from __future__ import annotations
import numpy as np

from typing import TYPE_CHECKING, Callable, Literal, Sequence, Tuple, overload

from ..typing import Dims, Coords
from ..basic_utils import permute, align, shapes_broadcastable, mergecoords, mergedims

from ._types import Broadcaster, Dimcaster, DimMerger, CoordMerger

from ._dimcast import unilateral_dimcast, trivial_dimcast

from ._template import Template


if TYPE_CHECKING:
    from .._base import XTensor



class TensorBroadcastError(ValueError):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


def broadcaster_wrapper(
        dimcast: Dimcaster,
        dimmerge: DimMerger, coordmerge: CoordMerger,
        shapecheck: bool=True) -> Broadcaster:

    def _broadcast(X: XTensor, Y: XTensor):
        return broadcast(X, Y, dimcast, dimmerge, coordmerge, shapecheck)
    return _broadcast


def broadcast(X: XTensor, Y: XTensor, 
        dimcast: Dimcaster, dimmerge: DimMerger, coordmerge: CoordMerger,
        shapecheck: bool=True
        ) -> Tuple[np.ndarray, np.ndarray, Dims, Coords]:
    '''
    Given two named tensors, return two numpy ND arrays of the same rank that
    can be broadcast together, along with the dimension names and coordinates
    after broadcasting.
    '''


    axes_x, axes_y = dimcast(X, Y)

    X1 = permute(X, axes_x)
    Y1 = permute(Y, axes_y)

    X1, Y1 = align(X1, Y1)

    newdims = dimmerge(X1, Y1)
    newcoords = coordmerge(X1, Y1)

    if shapecheck:
        if not shapes_broadcastable(X1.shape, Y1.shape):
            raise TensorBroadcastError(
                    f'Broadcast impossible with shapes and dims <{X.dims},{X.shape}> and <{Y.dims},{Y.shape}>')

    return X1.data, Y1.data, newdims, newcoords


unilateral_broadcaster = broadcaster_wrapper(
            dimcast=unilateral_dimcast(strict=False),
            dimmerge=mergedims, coordmerge=mergecoords
)
'''
This broadcaster takes the second tensor and permutes its dimensions to match
the first one. The algorithm for finding the permutation is defined by
unilateral_dimcast()
'''


vanilla_broadcaster = broadcaster_wrapper(
            dimcast=trivial_dimcast,
            dimmerge=mergedims, coordmerge=mergecoords
)
'''
Vanilla broadcaster: the same as what's used in torch's named tensors.
Dimension names (or None) have to match at the same axis position.
'''


def template_broadcaster(dims: Sequence[str|int], channels: Sequence[Literal['x','y']|None]) -> Broadcaster:
    '''
    Return a template broadcaster with the specified dimensions and channels
    '''

    template = Template.from_dims_channels(dims, channels)
    
    def _broadcast(X: XTensor, Y: XTensor):
        X1 = template.cast_and_update(X, 'x')
        Y1 = template.cast_and_update(Y, 'y')

        dims, coords = template.dims, template.coords
        template.clear()
        return X1.data, Y1.data, dims, coords

    # _broadcast.cast = lambda x, y: y

    return _broadcast


@overload
def cast(broadcaster: Broadcaster, X: XTensor) -> Callable[[XTensor], XTensor]: ...
@overload
def cast(broadcaster: Broadcaster, X: XTensor, Y: XTensor) -> XTensor: ...

def cast(broadcaster: Broadcaster, X: XTensor, Y: XTensor|None=None):
    if Y is None:
        def _cast(Y1: XTensor, /):
            x, _y, dims, coords = broadcaster(X, Y1)

            coords = [(coord if _y.shape[axis] > 1 else None) 
                    for axis, coord in enumerate(coords)]

            return Y1.__class__(_y, dims, coords)
        return _cast
    return cast(broadcaster, X)(Y)




from __future__ import annotations
from typing import TYPE_CHECKING, Any, List, Protocol, Sequence, Tuple
import numpy as np


if TYPE_CHECKING:
    from ...tensor._base import XTensor

from ..typing import AxesPermutation, Coords, Dims


class Dimcaster(Protocol):
    def __call__(self, X: XTensor, Y: XTensor) -> Tuple[AxesPermutation, AxesPermutation]: ...


class DimMerger(Protocol):
    def __call__(self, X: XTensor, Y: XTensor) -> Dims: ...


class CoordMerger(Protocol):
    def __call__(self, X: XTensor, Y: XTensor) -> Coords: ...


class Broadcaster(Protocol):
    def __call__(self,
            X: XTensor, Y: XTensor) -> Tuple[np.ndarray, np.ndarray, Dims, Coords]: ...




from __future__ import annotations
from typing import TYPE_CHECKING, Any, List, Sequence
import numpy as np
from torch import wait

from ..typing import Coords
if TYPE_CHECKING: from .._base import XTensor


def mergecoords(X: XTensor|Coords, Y: XTensor|Coords, rtol: float=1e-8, atol: float=1e-8) -> Coords:
    '''
    '''

    from .._base import XTensor

    if isinstance(X, XTensor): coords_x = list(X.coords)
    else: coords_x = X

    if isinstance(Y, XTensor): coords_y = list(Y.coords)
    else: coords_y = Y

    rank_x, rank_y = len(coords_x), len(coords_y)

    if rank_x < rank_y: return mergecoords(Y, X)


    newcoords: Coords = []

    coords_y: Coords = [None for _ in range(rank_y, rank_x)] + coords_y
    
    if not coords_same(coords_x, coords_y, rtol=rtol, atol=atol, none_compatible=True):
        raise ValueError('Coordinates incompatible')

    for coord_x, coord_y in zip(coords_x, coords_y):
        newcoords.append(coord_x if coord_x is not None else coord_y if coord_y is not None else None)

    return newcoords


def coords_same(coords1: Coords, coords2: Coords, /, *,
        rtol: float=1e-8, atol: float=1e-8, none_compatible: bool=False) -> bool:
    '''
    Return whether the given coordinates are the same.
    Two coordinate sequences are considered the same if:
        - 
    '''
    
    if len(coords1) != len(coords2): return False

    for coord1, coord2 in zip(coords1, coords2):
        if coord1 is None or coord2 is None: 
            condition = (coord1 is None and coord2 is None) or none_compatible
        else:
            try: 
                condition = np.allclose(coord1, coord2, rtol=rtol, atol=atol)
            except TypeError:
                condition = (coord1 == coord2)

        if not condition: return False

    return True



from __future__ import annotations
from typing import List, Literal, Optional, Sequence, Tuple, cast
import numpy as np
from xarray import DataArray

from xtensors.typing import DimLike, DimsLike, AxisDimPair
from xtensors.typing import NDArray


# def dims_to_axes(dims: Tuple[str,...], target_dims: Tuple[str,...]) -> Tuple[int,...]:
#     '''
#     Given a tuple of strings, find the indices where target_dims occur. 
#     '''
#     res: List[int] = []
#    
#     _dims = np.array(dims)
#
#     for tdim in target_dims:
#         index = np.where(_dims == tdim)[0]
#
#         index = index.item()
#         res.append(index)
#
#     return tuple(res)


def dim_to_axis(dims: Tuple[str,...], target_dim: str) -> int:
    '''
    Given a tuple of strings, find the index where target_dim occurs. If there
    is not exactly one occurrence, an IndexError is raised.

    '''
    _dims = np.array(dims)
    
    index = np.where(_dims == target_dim)[0]

    if len(index) == 0:
        raise IndexError(f'dim {target_dim} is not found')

    if len(index) > 1:
        raise IndexError(f'dim {target_dim} is ambiguous')

    index = index.item()
    return index


def get_axis_from_dim(x: NDArray, dim: str, fallback_axis: Optional[int]=None) -> int:
    '''
    Given a single dimension (string) and an array x, try locating the
    corresponding axis (int). If not found, the fallback is returned if
    provided, else an IndexError is raised.
    '''
    try:
        if isinstance(x, DataArray):
            return dim_to_axis(cast(Tuple[str,...], x.dims), dim)

        raise IndexError(f'trying to locate dimension {dim} but array/tensor is not named')

    except IndexError as e:
        if fallback_axis is not None:
            return fallback_axis
        
        raise IndexError(*e.args)


def get_axes(x: NDArray, dim: DimLike|DimsLike|None) -> Tuple[int,...]:
    '''
    Given dimensions (in strings) or axes (in integers), return a list of
    integers corresponding to the dimensions/axes

    Parameters:
        x: Input array
        dim: dimensions specifications, e.g.:
            2
            'N'
            (2, 'N') - 'N', or 2 if 'N' is not found
            (2, 3, 4)
            ('N', 'C', 'H')
            ((2, 'N'), (3, 'C'), (4, 'H'))

    '''

    if dim is None:
        return tuple([i for i in range(len(x.shape))])

    if isinstance(dim, int):
        return (dim,)

    if isinstance(dim, str):
        return (get_axis_from_dim(x, dim),)
    
    if isinstance(dim, tuple):
        if isinstance(dim[0], int):
            if isinstance(dim[1], str):
                return (get_axis_from_dim(x, dim[1], fallback_axis=dim[0]),)
            
            return cast(Tuple[int,...], dim)

        if isinstance(dim[0], str):
            dim = cast(Tuple[str,...], dim)
            return tuple(get_axis_from_dim(x, _dim) for _dim in dim)

        if isinstance(dim[0], tuple):
            dim = cast(Tuple[AxisDimPair,...], dim)
            return tuple(get_axis_from_dim(x, _dim[1], fallback_axis=_dim[0]) for _dim in dim)

    raise ValueError('Invalid argument combination')


def get_coord(x: NDArray, dim: str) -> np.ndarray | None:
    if isinstance(x, DataArray):
        return x.coords[dim].__array__()
    return None


def strip_dims(olddims: Tuple[str, ...], axis: Tuple[int, ...]) -> Tuple[str,...]:
    '''
    (For named dimensions) Remove dimensions corresponding to [axis] from named
    dimensions.
    '''
    newdims = list(olddims)
    stripped_dims = [olddims[i] for i in axis] 

    for sd in stripped_dims: newdims.remove(sd)

    return tuple(newdims)




def rename(x: NDArray, dim: DimLike, newname: str) -> DataArray:
    '''
    Rename a dimension. If the original array is unnamed, a named array will be
    returned.
    '''
    if isinstance(x, DataArray):
        _axis = get_axes(x, dim)[0]
        _dim = x.dims[_axis]
        return x.rename({_dim: newname})

    _x = x.__array__()
    _x = DataArray(_x)

    if isinstance(dim, int):
        return _x.rename({_x.dims[dim]: newname})

    if isinstance(dim, tuple):
        return _x.rename({_x.dims[dim[0]]: newname})

    raise ValueError


def new_axes(x: NDArray, *dims: str, position: Literal['before','after']='after') -> DataArray:
    _x = to_dataarray(x)
    if position == 'after':
        for dim in dims:
            _x = new_axis(_x, dim, position='after')
        return _x

    if position == 'before':
        for dim in dims[::-1]:
            _x = new_axis(_x, dim, position='before')
        return _x

    raise ValueError(f'Invalid position: {position}')


def new_axis(x: NDArray, dim: str, position: Literal['before','after']='after') -> DataArray:
    _x = to_dataarray(x)
    dims = list(_x.dims)
    _x = x.__array__()

    if position == 'before':
        _x = _x[None,...]
        _x = DataArray(_x, dims=['dims_appended_tmp'] + dims)
        _x = rename(_x, 0, dim)
        return _x

    if position == 'after':
        _x = _x[...,None]
        _x = DataArray(_x, dims=dims + ['dims_appended_tmp'])
        _x = rename(_x, -1, dim)
        return _x

    raise ValueError(f'Invalid position: {position}')



from __future__ import annotations

from typing import Any
import numpy as np
import numpy.typing as npt

from .. import tensor as xtt
from scipy import special



@xtt.generalize_at_2
@xtt.generalize_at_1
@xtt.generalize_at_0
@xtt.promote_ternary_operator()
def where(X: npt.NDArray, Y: npt.NDArray, Z: npt.NDArray) -> npt.NDArray:
    return np.where(X, Y, Z)


@xtt.generalize_at_0
def softmax(X: xtt.XTensor, /, dim: xtt.DimLike) -> xtt.XTensor:
    axis = X.get_axis(dim)
    _y = special.softmax(X.data, axis=axis)
    return xtt.XTensor(_y, dims=X.dims, coords=X.coords)


def get_rank(x: Any) -> int:
    if hasattr(x, 'shape'):
        return len(x.shape)

    if hasattr(x, '__array__'):
        return get_rank(x.__array__())

    if isinstance(x, (list, tuple)):
        return get_rank(x[0]) + 1

    return 0




from ._typing import AxesPermutation, Dims, Coords, BinaryOperator, Array, TernaryOperator, TensorLike
from ._typing import Function_2Args, Function_1Arg, Function_3Args, DimLike, DimsLike
from ._typing import HasDimName



from __future__ import annotations

from xtensors.functionals.base import Functional
from xtensors import base

from .. import tensor as xtt


class Reduction(Functional):
    '''
    Single-axis reduction fucntional
    '''
    def __init__(self, dim: xtt.DimLike, /, *args) -> None:
        self.dim = dim
        self._reduce: base._reduc.ReductionFunc | base._arg.ArgFunction
        self.name = 'UNIMPLEMENTED_REDUCTION'

    def __call__(self, x: xtt.TensorLike) -> xtt.XTensor:
        return self._reduce(x, self.dim)


class Index(Reduction):
    def __init__(self, dim: xtt.DimLike, index: int):
        self.dim = dim
        self.index = index
        self.name = f'Index({dim},{index})'

    @xtt.generalize_at_1
    def __call__(self, X: xtt.XTensor) -> xtt.XTensor:
        return xtt.index(X, (self.dim, self.index))


class Mean(Reduction):
    def __init__(self, dim: int|str, nan: bool=True) -> None:
        super().__init__(dim)
        self._reduce = base.nanmean if nan else base.mean
        self.name = f'Mean({dim})'


class Sum(Reduction):
    def __init__(self, dim: int|str, nan: bool=True) -> None:
        super().__init__(dim)
        self._reduce = base.nansum if nan else base.sum
        self.name = f'Sum({dim})'


class Std(Reduction):
    def __init__(self, dim: int|str, nan: bool=True) -> None:
        super().__init__(dim)
        self._reduce = base.nanstd if nan else base.std
        self.name = f'Std({dim})'


class Max(Reduction):
    def __init__(self, dim: int|str, nan: bool=True) -> None:
        super().__init__(dim)
        self._reduce = base.nanmax if nan else base.max
        self.name = f'Max({dim})'


class Min(Reduction):
    def __init__(self, dim: int|str, nan: bool=True) -> None:
        super().__init__(dim)
        self._reduce = base.nanmin if nan else base.min
        self.name = f'Max({dim})'


class ArgMax(Reduction):
    def __init__(self, dim: int|str, nan: bool=True) -> None:
        super().__init__(dim)
        self._reduce = base.nanargmax if nan else base.argmax
        self.name = f'ArgMax({dim})'


class ArgMin(Reduction):
    def __init__(self, dim: int|str, nan: bool=True) -> None:
        super().__init__(dim)
        self._reduce = base.nanargmin if nan else base.argmin
        self.name = f'ArgMin({dim})'


class CoordMax(Reduction):
    def __init__(self, dim: int|str, nan: bool=True) -> None:
        super().__init__(dim)
        self._reduce = base.nancoordmax if nan else base.coordmax
        self.name = f'CoordMax({dim})'


class CoordMin(Reduction):
    def __init__(self, dim: int|str, nan: bool=True) -> None:
        super().__init__(dim)
        self._reduce = base.nancoordmin if nan else base.coordmin
        self.name = f'CoordMin({dim})'



from __future__ import annotations
from typing import Protocol

import numpy as np

from .. import tensor as xtt


class _ufunc(Protocol):
    def __call__(self, __x1: np.ndarray, /, *args) -> np.ndarray: ...


class UFunc(Protocol):
    def __call__(self, x: xtt.TensorLike, /) -> xtt.XTensor: ...


def _ufunc_factory(_np_func: _ufunc) -> UFunc:
    @xtt.generalize_at_0
    def _f(X: xtt.XTensor, /) -> xtt.XTensor:
        return xtt.XTensor(_np_func(X.data), X.dims, X.coords)
    return _f


def _sigmoid(__x1: np.ndarray):
    return .5 * (1. + np.tanh(__x1/2))

sigmoid = _ufunc_factory(_sigmoid)

exp = _ufunc_factory(np.exp)

log2 = _ufunc_factory(np.log2)
log = _ufunc_factory(np.log)
log10 = _ufunc_factory(np.log10)

cos = _ufunc_factory(np.cos)
sin = _ufunc_factory(np.sin)
tan = _ufunc_factory(np.tan)

cosh = _ufunc_factory(np.cosh)
sinh = _ufunc_factory(np.sinh)
tanh = _ufunc_factory(np.tanh)




from __future__ import annotations
from typing import TYPE_CHECKING, List, Literal, Sequence

from ... import numpy as xtnp

from ._axes import permute

from ._generalize import generalize_at_0


if TYPE_CHECKING:
    from .._base import XTensor
    from ..typing import DimsLike, DimLike, Dims


def mergedims(X: XTensor|Dims, Y: XTensor|Dims) -> Dims:
    '''
    '''

    from .._base import XTensor

    if isinstance(X, XTensor): dims_x = list(X.dims)
    else: dims_x = X

    if isinstance(Y, XTensor): dims_y = list(Y.dims)
    else: dims_y = Y


    rank_x, rank_y = len(dims_x), len(dims_y)

    if rank_x < rank_y: return mergedims(Y, X)

    newdims = []

    dims_y = [None for _ in range(rank_y, rank_x)] + dims_y
    
    for dim_x, dim_y in zip(dims_x, dims_y):
        if dim_x is not None and dim_y is not None and dim_x != dim_y:
            raise ValueError(f'Incompatible dimensions: {dims_x} and {dims_y}')
        newdims.append(dim_x if dim_x is not None else dim_y)

    return newdims


def dimslast(X: XTensor, dims: Sequence[str]) -> XTensor:
    '''Move named dimensions to the right.'''
    axes: List[int] = [X.get_axis(dim) for dim in dims]
    other_axes: List[int|None] = [axis for axis in range(X.rank) if axis not in axes]

    return permute(X, other_axes+axes)


def dimsfirst(X: XTensor, dims: Sequence[str]) -> XTensor:
    '''Move named dimensions to the left.'''
    axes: List[int] = [X.get_axis(dim) for dim in dims]
    other_axes: List[int|None] = [axis for axis in range(X.rank) if axis not in axes]

    return permute(X, axes+other_axes)

@generalize_at_0
def flatten(X: XTensor, /,
        dims: DimsLike, dim_out: str|None, *,
        position: Literal['left', 'right']='right') -> XTensor:

    from .._base import XTensor
    x = X.data
    # axes = [X.get_axis(dim) for dim in dims]
    axes = X.get_axes(dims)

    remaining_dims = [dim for axis, dim in enumerate(X.dims) if axis not in axes]
    remaining_coords = [coord for axis, coord in enumerate(X.coords) if axis not in axes]

    # TODO: implement coordinate meshgrid
    coord_out = None

    x_flat = xtnp.flatten(x, axes, position=position)

    if position == 'left':
        Y = XTensor(x_flat, [dim_out]+remaining_dims, [coord_out] + remaining_coords)
        return Y

    else:
        Y = XTensor(x_flat, remaining_dims+[dim_out], remaining_coords+[coord_out])
        return Y


@generalize_at_0
def name_dim_if_absent(X: XTensor, /, axis: int, dim: str, *, force: bool=False) -> XTensor:
    '''
    Ensure that X has a named dimension called [dim]. If not already, the
    dimension at [axis] will be named so.
    '''
    X1 = X.viewcopy()
    if dim in X.dims:
        return X1

    if X1.dims[axis] is not None and not force:
        raise ValueError(f'Tensor already has its dimension named {X1.dims[axis]} at axis={axis}')

    X1.set_dim(axis, dim)
    return X1


@generalize_at_0
def dims(X: XTensor): return X.dims




from __future__ import annotations
'''
Functionalities with np arrays
'''
from typing import List, Literal, Sequence, Union, cast
import numpy as np
from numpy.typing import NDArray


AxesLike = Union[int,Sequence[int],None]


def _axes_list(a: np.ndarray, axes: AxesLike) -> List[int]:
    if axes is None: return [i for i in range(len(a.shape))]

    if isinstance(axes, int): return [axes]

    return list(axes)


def axes_last(a: np.ndarray, axes: AxesLike=None) -> np.ndarray:
    '''
    Reshape the array so that the specified axes are placed at the end
    '''
    axes = _axes_list(a, axes)
    newaxes = [i for i in range(len(a.shape)) if i not in axes] + axes

    return a.transpose(*newaxes)


def axes_first(a: np.ndarray, axes: AxesLike=None) -> np.ndarray:
    '''
    Reshape the array so that the specified axes are placed at the beginning (left)
    '''
    axes = _axes_list(a, axes)
    newaxes = axes + [i for i in range(len(a.shape)) if i not in axes]

    return a.transpose(*newaxes)


def flatten(a: np.ndarray, axes: AxesLike=None, position: Literal['left', 'right']='right') -> np.ndarray:
    axes = _axes_list(a, axes)
    n_preserved_axes = len(a.shape) - len(axes)
    

    if position == 'right':
        b = axes_last(a, axes)
        return b.reshape(*b.shape[:n_preserved_axes], -1)
    else:
        b = axes_first(a, axes)
        return b.reshape(-1, *b.shape[-n_preserved_axes:])


def fold_indices(indices: NDArray[np.int_], fold_shape: Sequence[int]) -> NDArray[np.int_]:
    '''
        indices: (*)
        fold_shape: [m1, m2, m3, ..., mN]

        return: folded indices (*, N)
    '''
    _indices = []
    indices = indices.copy()

    for l in fold_shape[::-1]:
        _indices.append(indices % l)
        indices //= l

    return np.array(_indices[::-1]).transpose(*[i+1 for i in range(len(indices.shape))], 0)
        


def bincount(x: NDArray[np.int_], N: int|None=None) -> NDArray[np.int_]:
    '''
        x: (*, M)
        return: (*, N)
    '''

    if N is None:
        N = np.max(x) + 1

    shape = x.shape[:-1]

    # (D, M)
    _x = flatten(x, [a for a in range(len(x.shape))][:-1], position='left')

    D = _x.shape[0]
    index = np.arange(D).reshape(D, 1)


    N = cast(int, N)

    y = np.bincount((index*N + _x).reshape(-1), minlength=N*D).reshape(*shape, N)

    return y



    
    



from ..typing import NDArray

from xarray import DataArray
import numpy as np

from ..tensor._base import XTensor

def generate_ignored_dims(n: int):
    return [f'__dim{i}' for i in range(n)]


def to_dataarray(x: NDArray) -> DataArray:
    if isinstance(x, DataArray): return x
    return DataArray(x, dims=generate_ignored_dims(len(x.shape)))


def to_xtensor(x: NDArray) -> XTensor:
    if isinstance(x, XTensor): return x

    return XTensor(x.__array__(), dims=generate_ignored_dims(len(x.shape)))


def to_ndarray(x: NDArray) -> np.ndarray:
    return x.__array__()



from __future__ import annotations

from .. import tensor as xtt

from typing import Optional, Protocol

import warnings


class BinaryOperation(Protocol):
    def __call__(self, x: xtt.TensorLike, y: xtt.TensorLike, /) -> xtt.XTensor: ...


def _apply_operation(X: xtt.XTensor, Y: xtt.XTensor, binop: str, rbinop: Optional[str]=None) -> xtt.XTensor:

    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', 'invalid value encountered in true_divide')
        try:
            Z = getattr(X, binop)(Y)
            if Z is NotImplemented: raise NotImplementedError

        except (AttributeError, NotImplementedError) as e:
            if rbinop is not None:
                Z = getattr(Y, rbinop)(X)
                if Z is NotImplemented: raise NotImplementedError
            else:
                raise AttributeError from e
    return Z


def _binop_factory(_bin_op: str, _rbin_op: str) -> BinaryOperation:
    @xtt.generalize_at_1
    @xtt.generalize_at_0
    def _op(X: xtt.XTensor, Y: xtt.XTensor, /) -> xtt.XTensor:
        return _apply_operation(X, Y, _bin_op, _rbin_op)

    return _op


_add = _binop_factory('__add__', '__radd__')
_divide = _binop_factory('__truediv__', '__rtruediv__')
_multiply = _binop_factory('__mul__', '__rmul__')

_greater = _binop_factory('__gt__', '__lt__')
_greater_equal = _binop_factory('__ge__', '__le__')

_less = _binop_factory('__lt__', '__gt__')
_less_equal = _binop_factory('__le__', '__ge__')

_equal = _binop_factory('__eq__', '__eq__')





from __future__ import annotations
from typing import List, Literal, Sequence, Tuple, TypeVar, overload



T = TypeVar('T')

@overload
def strip(l: Sequence[T], indices: Sequence[int]) -> List[T]: ...
@overload
def strip(l: Sequence[T], indices: Sequence[int], only_remaining: Literal[True]) -> List[T]: ...
@overload
def strip(l: Sequence[T], indices: Sequence[int], only_remaining: Literal[False]) -> Tuple[List[T], List[T]]: ...

def strip(l: Sequence[T], indices: Sequence[int], only_remaining: bool=True) -> List[T]|Tuple[List[T], List[T]]:
    remaining, stripped = [], []

    for i, x in enumerate(l):
        if i in indices: stripped.append(x)
        else: remaining.append(x)

    if only_remaining: return remaining
    return remaining, stripped




